@file:Suppress("unused")
@file:JvmName("SafeAutocloseKt")

package ru.landgrafhomyak.utility.highlevel_try_finally

import kotlin.contracts.InvocationKind
import kotlin.contracts.contract
import kotlin.jvm.JvmName

@Suppress("WRONG_INVOCATION_KIND")
inline fun <R> safeAutoClose1(
	finally: () -> Unit,
	action: () -> R
): R {
	contract {
		callsInPlace(action, InvocationKind.EXACTLY_ONCE)
		callsInPlace(finally, InvocationKind.EXACTLY_ONCE)
	}
	return safeAutoClose3(onError = finally, onSuccess = finally, onCrossReturn = finally, action = action)
}

@Suppress("WRONG_INVOCATION_KIND")
inline fun <R> safeAutoClose2(
	onError: () -> Unit = {},
	onSuccess: () -> Unit = {},
	action: () -> R
): R {
	contract {
		callsInPlace(action, InvocationKind.EXACTLY_ONCE)
		callsInPlace(onError, InvocationKind.AT_MOST_ONCE)
		callsInPlace(onSuccess, InvocationKind.AT_MOST_ONCE)
	}
	return safeAutoClose3(onError = onError, onSuccess = onSuccess, onCrossReturn = onSuccess, action = action)
}

@Suppress("WRONG_INVOCATION_KIND")
inline fun <R> safeAutoClose2e(
	onError: (Throwable) -> Unit = { _ -> },
	onSuccess: () -> Unit = {},
	action: () -> R
): R {
	contract {
		callsInPlace(action, InvocationKind.EXACTLY_ONCE)
		callsInPlace(onError, InvocationKind.AT_MOST_ONCE)
		callsInPlace(onSuccess, InvocationKind.AT_MOST_ONCE)
	}
	return safeAutoClose3e(onError = onError, onSuccess = onSuccess, onCrossReturn = onSuccess, action = action)
}

inline fun <R> safeAutoClose3(
	onError: () -> Unit = {},
	onSuccess: () -> Unit = {},
	onCrossReturn: () -> Unit = {},
	action: () -> R
): R {
	contract {
		callsInPlace(action, InvocationKind.EXACTLY_ONCE)
		callsInPlace(onError, InvocationKind.AT_MOST_ONCE)
		callsInPlace(onSuccess, InvocationKind.AT_MOST_ONCE)
		callsInPlace(onCrossReturn, InvocationKind.AT_MOST_ONCE)
	}
	return safeAutoClose3e(onError = { t -> onError() }, onSuccess = onSuccess, onCrossReturn = onCrossReturn, action = action)
}

inline fun <R> safeAutoClose3e(
	onError: (Throwable) -> Unit = { _ -> },
	onSuccess: () -> Unit = {},
	onCrossReturn: () -> Unit = {},
	action: () -> R
): R {
	contract {
		callsInPlace(action, InvocationKind.EXACTLY_ONCE)
		callsInPlace(onError, InvocationKind.AT_MOST_ONCE)
		callsInPlace(onSuccess, InvocationKind.AT_MOST_ONCE)
		callsInPlace(onCrossReturn, InvocationKind.AT_MOST_ONCE)
	}

	val ret: R
	var wasError = false
	var crossReturned = true
	try {
		ret = action()
		crossReturned = false
	} catch (e1: Throwable) {
		wasError = true
		try {
			onError(e1)
		} catch (e2: Throwable) {
			e1.addSuppressed(e2)
		}
		throw e1
	} finally {
		if (!wasError) {
			if (crossReturned)
				onCrossReturn()
			else
				onSuccess()
		}
	}
	return ret
}